# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0
import logging

from typing import Any, ClassVar, Dict, Union

from dataclasses import dataclass, field

from etptypes.energistics.etp.v12.protocol.core.open_session import OpenSession
from etptypes.energistics.etp.v12.protocol.core.request_session import (
    RequestSession,
)

from etpproto.endpoint_capability_kind import kind_from_name


@dataclass
class ClientInfo:
    count_instance: ClassVar[int] = 0

    endpoint_capabilities: Dict[str, Any] = field(
        default_factory=lambda: {
            "MaxWebSocketFramePayloadSize": 10000,
            "MaxWebSocketMessagePayloadSize": 10000,
        }
    )
    login: str = field(default="anonymousUser")
    ip: str = field(default="0.0.0.0")
    authenticated: bool = field(default=False)

    def __post_init__(self):
        self._id = self.count_instance
        ClientInfo.count_instance = ClientInfo.count_instance + 1

    def getCapability(self, name: str) -> Any:
        if name in self.endpoint_capabilities:
            return self.endpoint_capabilities[name]
        return None

    def negotiate(self, msg: Union[OpenSession, RequestSession]):
        for k, v in msg.endpoint_capabilities.items():
            cap_class = kind_from_name(k)
            # if cap_class != None:
            val: Any = v
            if hasattr(val, "item"):
                val = v.item  # if it is a DataValue

            if k in self.endpoint_capabilities:
                try:
                    val = min(val, self.endpoint_capabilities[k])
                except Exception as e1:
                    logging.debug(e1)

            if cap_class is not None:
                if cap_class._min is not None:
                    val = max(val, cap_class._min)
                if cap_class._max is not None:
                    val = min(val, cap_class._max)

            self.endpoint_capabilities[k] = val

        logging.debug(f"Negotiated capa : {self.endpoint_capabilities}")
        # else:
        # logging.debug(f"No capability found for name '{k}'")

    def __str__(self) -> str:
        return (
            "ClientInfo["
            + str(self._id)
            + "] "
            + self.ip
            + "<"
            + str(self.login)
            + ">"
        )

    def __hash__(self) -> int:
        return self._id

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ClientInfo):
            return NotImplemented
        return self._id == other._id



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

from typing import AsyncGenerator, Optional, Protocol, Union

from etptypes.energistics.etp.v12.datatypes.message_header import MessageHeader
from etptypes.energistics.etp.v12.protocol.channel_subscribe.cancel_get_ranges import (
    CancelGetRanges,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.channel_data import (
    ChannelData,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.channels_truncated import (
    ChannelsTruncated,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.get_change_annotations import (
    GetChangeAnnotations,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.get_change_annotations_response import (
    GetChangeAnnotationsResponse,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.get_channel_metadata import (
    GetChannelMetadata,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.get_channel_metadata_response import (
    GetChannelMetadataResponse,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.get_ranges import (
    GetRanges,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.get_ranges_response import (
    GetRangesResponse,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.range_replaced import (
    RangeReplaced,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.subscribe_channels import (
    SubscribeChannels,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.subscribe_channels_response import (
    SubscribeChannelsResponse,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.subscriptions_stopped import (
    SubscriptionsStopped,
)
from etptypes.energistics.etp.v12.protocol.channel_subscribe.unsubscribe_channels import (
    UnsubscribeChannels,
)

from etpproto.client_info import ClientInfo
from etpproto.error import InvalidMessageTypeError, NotSupportedError
from etpproto.messages import Message
from etpproto.utils import snake_case


class ChannelSubscribeHandler(Protocol):
    async def on_cancel_get_ranges(
        self,
        msg: CancelGetRanges,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_channel_data(
        self,
        msg: ChannelData,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_channels_truncated(
        self,
        msg: ChannelsTruncated,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_change_annotations(
        self,
        msg: GetChangeAnnotations,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_change_annotations_response(
        self,
        msg: GetChangeAnnotationsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_channel_metadata(
        self,
        msg: GetChannelMetadata,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_channel_metadata_response(
        self,
        msg: GetChannelMetadataResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_ranges(
        self,
        msg: GetRanges,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_ranges_response(
        self,
        msg: GetRangesResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_range_replaced(
        self,
        msg: RangeReplaced,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_subscribe_channels(
        self,
        msg: SubscribeChannels,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_subscribe_channels_response(
        self,
        msg: SubscribeChannelsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_subscriptions_stopped(
        self,
        msg: SubscriptionsStopped,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_unsubscribe_channels(
        self,
        msg: UnsubscribeChannels,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def handle_message(
        self,
        etp_object: object,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        handling_func = getattr(
            self, "on_" + snake_case(type(etp_object).__name__)
        )
        if handling_func is not None:
            async for handled in handling_func(
                msg=etp_object,
                msg_header=msg_header,
                client_info=client_info,
            ):
                yield handled

        else:
            raise InvalidMessageTypeError()



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

from typing import AsyncGenerator, Optional, Protocol, Union

from etptypes.energistics.etp.v12.datatypes.message_header import MessageHeader
from etptypes.energistics.etp.v12.protocol.store.chunk import Chunk
from etptypes.energistics.etp.v12.protocol.store.delete_data_objects import (
    DeleteDataObjects,
)
from etptypes.energistics.etp.v12.protocol.store.delete_data_objects_response import (
    DeleteDataObjectsResponse,
)
from etptypes.energistics.etp.v12.protocol.store.get_data_objects import (
    GetDataObjects,
)
from etptypes.energistics.etp.v12.protocol.store.get_data_objects_response import (
    GetDataObjectsResponse,
)
from etptypes.energistics.etp.v12.protocol.store.put_data_objects import (
    PutDataObjects,
)
from etptypes.energistics.etp.v12.protocol.store.put_data_objects_response import (
    PutDataObjectsResponse,
)

from etpproto.client_info import ClientInfo
from etpproto.error import InvalidMessageTypeError, NotSupportedError
from etpproto.messages import Message
from etpproto.utils import snake_case


class StoreHandler(Protocol):
    async def on_chunk(
        self,
        msg: Chunk,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_delete_data_objects(
        self,
        msg: DeleteDataObjects,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_delete_data_objects_response(
        self,
        msg: DeleteDataObjectsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_data_objects(
        self,
        msg: GetDataObjects,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_data_objects_response(
        self,
        msg: GetDataObjectsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_data_objects(
        self,
        msg: PutDataObjects,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_data_objects_response(
        self,
        msg: PutDataObjectsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def handle_message(
        self,
        etp_object: object,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        handling_func = getattr(
            self, "on_" + snake_case(type(etp_object).__name__)
        )
        if handling_func is not None:
            async for handled in handling_func(
                msg=etp_object,
                msg_header=msg_header,
                client_info=client_info,
            ):
                yield handled

        else:
            raise InvalidMessageTypeError()



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

from typing import ClassVar, Optional

from etptypes.energistics.etp.v12.datatypes.error_info import ErrorInfo
from etptypes.energistics.etp.v12.protocol.core.protocol_exception import (
    ProtocolException,
)

from etpproto.messages import Message


# https://stackoverflow.com/a/25004536


class ETPError(Exception):
    """Base class for ETP exceptions"""

    code: ClassVar[int] = 0

    def __init__(self, msg: str) -> None:
        super().__init__()
        self.msg = msg

    def to_etp_error(self) -> ErrorInfo:
        return ErrorInfo(message=str(self.msg), code=self.code)
        # return ErrorInfo(message=str(self), code=self.code)

    def to_etp_message(
        self, msg_id: int = -1, correlation_id: int = 0
    ) -> Optional[Message]:
        err_info = self.to_etp_error()
        return Message.get_object_message(
            ProtocolException(error=err_info, errors={}),
            msg_id=msg_id,
            correlation_id=correlation_id,
        )


class NoRoleError(ETPError):
    code: ClassVar[int] = 1

    def __init__(self) -> None:
        super().__init__("The endpoint does not support the requested role.")


class NoSupportedProtocolsError(ETPError):
    code: ClassVar[int] = 2

    def __init__(self) -> None:
        super().__init__(
            "The server does not support any of the requested protocols."
        )


class InvalidMessageTypeError(ETPError):
    code: ClassVar[int] = 3

    def __init__(self) -> None:
        super().__init__(
            "The message type ID is either: 1) not defined at all in the ETP Specification (e.g., no schema for it); or 2) not a correct message type ID for the receiving role (EXAMPLE: Per this specification, only the store role may SEND a GetDataObjectsResponse message; if the store RECEIVES a GetDataObjectsResponse message, it MUST send this error code.)",
        )


class UnsupportedProtocolError(ETPError):
    code: ClassVar[int] = 4

    def __init__(self, protocol_id: int) -> None:
        super().__init__(
            f"The endpoint does not support the protocol (#{protocol_id}) identified in a message header.",
        )


class InvalidArgumentError(ETPError):
    """Use this error code in any situation where a logically invalid argument is encountered."""

    code: ClassVar[int] = 5

    def __init__(self) -> None:
        super().__init__("Logically invalid argument.")


class RequestDeniedError(ETPError):
    """
    RECOMMENDATION: Endpoints should supply an error message explaining why the request was denied.
    For example, for read-only servers (which do not support Store operations),
    the explanation could be "Read-only server; operation not supported.
    """

    code: ClassVar[int] = 6

    def __init__(self, more: str) -> None:
        super().__init__("Server has denied the request. " + more)


class NotSupportedError(ETPError):
    code: ClassVar[int] = 7

    def __init__(self) -> None:
        super().__init__("The operation is not supported by the endpoint.")


class InvalidStateError(ETPError):
    """
    Indicates that the message is not allowed in the current state of the protocol.
    For example, sending ChannelStreamingStart for a channel that is already streaming,
    or received a message that is not applicable for the current role.
    """

    code: ClassVar[int] = 8

    def __init__(self) -> None:
        super().__init__(
            "The message is not allowed in the current state of the protocol.",
        )


class InvalidUriError(ETPError):
    """
    EXAMPLE: If a customer sends an alternate URI format to a store that does not accept/support
    alternate URIs, the store MUST send this error code.
    """

    code: ClassVar[int] = 9

    def __init__(self) -> None:
        super().__init__(
            "The URI sent is either a malformed URI or is not a valid URI format for ETP.",
        )


class ExpiredTockenError(ETPError):
    """
    Sent from server to client when the server is about to terminate the session because of an expired security token.
    """

    code: ClassVar[int] = 10

    def __init__(self) -> None:
        super().__init__("The security token is expired.")


class NotFoundError(ETPError):
    """
    Used when a resource (i.e., an object, part or range) is not found.
    """

    code: ClassVar[int] = 11

    def __init__(self) -> None:
        super().__init__("Resource not found.")


class LimitExceededError(ETPError):
    """
    Sent by a store if a request exceeds allowed limits or what the endpoint can handle. For example, this error code is used:
        - In Protocol 3 (Discover) and all query protocols, if the results of a client request exceeds the MaxResponseCount variable (which indicates the maximum number of resources a store will return).
        - In Protocol 21 (ChannelSubscribe) if a producer exceeds a consumers MaxDataItemCount.
    """

    code: ClassVar[int] = 12

    def __init__(self) -> None:
        super().__init__("Request exceeds allowed limits.")


class CompressionNotSupportedError(ETPError):
    """
    Sent by any role (producer, consumer, etc.) when it receives one of the following message types, which can never be compressed: RequestSession, OpenSession, ProtocolException or Acknowledge.
    """

    code: ClassVar[int] = 13

    def __init__(self) -> None:
        super().__init__("Message can not be compressed.")


class InvalidObjectError(ETPError):
    """
    Sent in any protocol when either role sends an invalid XML document. Note: ETP does not distinguish between well-formed and invalid for this purpose. The same error message is used in both cases.
    """

    code: ClassVar[int] = 14

    def __init__(self) -> None:
        super().__init__("Invalid XML document.")


class MaxTransactionsExceededError(ETPError):
    """
    The maximum number of transactions per ETP session has been exceeded. Currently, Transaction (Protocol 18) is the only ETP protocol that has the notion of a "transaction" and allows only 1 transaction per session.
    """

    code: ClassVar[int] = 15

    def __init__(self) -> None:
        super().__init__(
            "Maximum number of transactions per ETP session has been exceeded.",
        )


class ContentTypeNotSupportedError(ETPError):
    """
    The content type is not supported by the server.
    """

    code: ClassVar[int] = 16

    def __init__(self) -> None:
        super().__init__("The content type is not supported by the server.")


class MaxSizeExceededError(ETPError):
    """
    Sent from a store to a customer when the customer attempts a get or put operation that exceeds the stores maximum advertised MaxDataObjectSize, MaxPartSize, or MaxDataArraySize.
    """

    code: ClassVar[int] = 17

    def __init__(self) -> None:
        super().__init__(
            "Operation exceeds the stores maximum advertised MaxDataObjectSize, MaxPartSize, or MaxDataArraySize.",
        )


class MultipartCancelledError(ETPError):
    """
    Sent by either role to notify of canceled transmission of multi-message response or request when one of the maximum advertised protocol capabilities (maxConcurrentMultipart, maxMultipartMessageTimeInterval, or maxMultipartTotalSize) has been exceeded.
    """

    code: ClassVar[int] = 18

    def __init__(self) -> None:
        super().__init__("Canceled transmission of multi-message response.")


class InvalidMessageError(ETPError):
    """
    Sent by either endpoint when it is unable to deserialize the header or body of a message.
    """

    code: ClassVar[int] = 19

    def __init__(self) -> None:
        super().__init__(
            "Unable to deserialize the header or body of a message."
        )


class InvalidIndexKindError(ETPError):
    """
    Sent by either role when an IndexKind used in a message is invalid for the dataset. For example, see the Replace Range message in ChannelDataLoad (Protocol 22).
    """

    code: ClassVar[int] = 20

    def __init__(self) -> None:
        super().__init__(
            "IndexKind used in message is invalid for the dataset."
        )


class NoSupportedFormatsError(ETPError):
    """
    Sent by either role if, during session negotiation, no agreement can be reached on the format (XML or JSON) of data objects. The role that sends this message should then send the CloseSession message.
    """

    code: ClassVar[int] = 21

    def __init__(self) -> None:
        super().__init__(
            "No agreement can be reached on the format (XML or JSON) of data objects.",
        )


class RequestUuidRejectedError(ETPError):
    """
    Sent by the store when it rejects a customer-assigned request UUID (requestUuid), most likely because the request UUID is not unique within the session.
    """

    code: ClassVar[int] = 22

    def __init__(self) -> None:
        super().__init__(
            "Rejects a customer-assigned request UUID (requestUuid)."
        )


class UpdateGrowingObjectDeniedError(ETPError):
    """
    Sent by a store when a customer tries to update an existing growing object (i.e., do a put operation) using Store (Protocol 4). Growing objects can only be updated using GrowingObject (Protocol 6).
    """

    code: ClassVar[int] = 23

    def __init__(self) -> None:
        super().__init__(
            "Tryed to update an existing growing object using Store (Protocol 4).",
        )


class BackPressureLimitExceededError(ETPError):
    """
    Indicates the sender has detected the receiver is not processing messages as fast as it can send them and exceeding its capacity in its outgoing buffers. If sender capacity is exhausted and it is eas
    """

    code: ClassVar[int] = 24

    def __init__(self) -> None:
        super().__init__("Receiver's outgoing buffers capacity exceeded.")


class BackPressureWarningError(ETPError):
    """
    Listen to recording; at what criteria do you send the warming.
    """

    code: ClassVar[int] = 25

    def __init__(self) -> None:
        super().__init__("Back Pressure Warning.")


class TimedOut(ETPError):
    """
    May be sent by either role to cancel an operation when the response time for a relevant operation is exceeded, such as ResponseTimeoutPeriod or MultipartMessageTimeoutPeriod capabilities.
    """

    code: ClassVar[int] = 26

    def __init__(self) -> None:
        super().__init__("Response timeout")


class AuthorizationRequired(ETPError):
    """
    Sent from an endpoint during session negotiation (and ONLY during session negotiation) to indicate that the other endpoint requires authorization.
    """

    code: ClassVar[int] = 27

    def __init__(self) -> None:
        super().__init__("Authorization required.")


class AuthorizationExpiring(ETPError):
    """
        Optionally sent from an endpoint when the other endpoint's authorization will expire soon. The receiving endpoint should follow the necessary authorization workflow to renew its authorization. If it does not, the sending endpoint will eventually terminate the connection.
    The precise definition of "soon" and the required re-authorization workflow are intentionally out of the scope of the ETP Specification.
    """

    code: ClassVar[int] = 28

    def __init__(self) -> None:
        super().__init__("Authorization expiring.")


class NoSupportedDataObjectTypes(ETPError):
    """
    The server does not support any of the client's supported data object types.
    """

    code: ClassVar[int] = 29

    def __init__(self) -> None:
        super().__init__(
            "The server does not support any of the client's supported data object types."
        )


class ResponseCountExceeded(ETPError):
    """
    Sent by a store endpoint to terminate a non-map response once the number of responses sent has reached the allowed or stated limits specified by the relevant capabilities. This lets customers know that the store has more data than it could return to the customer. EXAMPLES:
        - In Protocol 3 (Discovery) and all query protocols, sent by the store if it must stop sending responses to the customer because it has already sent MaxResponseCount responses to a customer request.
        - In Protocol 21 (ChannelSubscribe), sent by the store if it must stop sending data points to a customer in response to a GetRanges request because the store has already sent MaxRangeDataItemCount data points in response to the request.
    """

    code: ClassVar[int] = 30

    def __init__(self) -> None:
        super().__init__("Response count exceeded.")


class InvalidAppend(ETPError):
    """
    Sent in response to a ChannelData message that is not appending data to a channel.
    """

    code: ClassVar[int] = 31

    def __init__(self) -> None:
        super().__init__(
            "Sent in response to a ChannelData message that is not appending data to a channel."
        )


class InvalidOperation(ETPError):
    """
    Sent in response to a request when the requested operation would be invalid. EXAMPLE: In Protocol 6 (GrowingObject), a ReplacePartsByRange message where some replacement parts are not covered by the delete range is an invalid operation.
    """

    code: ClassVar[int] = 32

    def __init__(self) -> None:
        super().__init__("Invalid operation.")


class InvalidChannelIDError(ETPError):
    """
    Sent by either role  when operations are requested on a channel that does not exist.
    """

    code: ClassVar[int] = 1002

    def __init__(self) -> None:
        super().__init__(
            "Operations are requested on a channel that does not exist."
        )


class UnsupportedObjectError(ETPError):
    """
    Sent in the Store protocols, when either role sends or requests a data object type that is not supported, according to the protocol capabilities.
    """

    code: ClassVar[int] = 4001

    def __init__(self) -> None:
        super().__init__(
            "Operations are requested on a channel that does not exist."
        )


class NoCascadeDeleteError(ETPError):
    """
    Sent when an attempt is made to delete an object that has children and the store does not support cascading deletes.
    """

    code: ClassVar[int] = 4003

    def __init__(self) -> None:
        super().__init__("Store does not support cascading deletes.")


class PluralObjectError(ETPError):
    """
    Sent when an endpoint uses puts for more than one object under the plural root of a 1.x Energistics data object. ETP only supports a single data object, one XML document.
    """

    code: ClassVar[int] = 4004

    def __init__(self) -> None:
        super().__init__(
            "ETP only supports a single data object, one XML document."
        )


class GrowingPortionIgnoredError(ETPError):
    """
    Sent from a store to a customer when the customer supplies the growing portion in a Put. This is advisory only; the object is upserted, but the growing portion is ignored.
    """

    code: ClassVar[int] = 4005

    def __init__(self) -> None:
        super().__init__("Customer supplies the growing portion in a Put.")


class RetentionPeriodExceededError(ETPError):
    """
    Sent from a store to a customer when the client asks for changes beyond the stated change period of a server.
    """

    code: ClassVar[int] = 5001

    def __init__(self) -> None:
        super().__init__(
            "Ask for changes beyond the stated change period of a server.",
        )


class NotGrowingObjectError(ETPError):
    """
    Sent from a store to a customer when the customer attempts to perform a growing object operation on an object that is not defined as a growing object type. This message does NOT apply to an object declared as a growing object but that is simply not actively growing at the present time.
    """

    code: ClassVar[int] = 6001

    def __init__(self) -> None:
        super().__init__(
            "Growing object operation on an object that is not defined as a growing object type.",
        )


class InternalError(ETPError):
    """
    Sent when an error occured but is was not an ETP specific error
    """

    code: ClassVar[int] = -1



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

from typing import AsyncGenerator, Optional, Protocol, Union

from etptypes.energistics.etp.v12.datatypes.message_header import MessageHeader
from etptypes.energistics.etp.v12.protocol.store_notification.chunk import (
    Chunk,
)
from etptypes.energistics.etp.v12.protocol.store_notification.object_access_revoked import (
    ObjectAccessRevoked,
)
from etptypes.energistics.etp.v12.protocol.store_notification.object_active_status_changed import (
    ObjectActiveStatusChanged,
)
from etptypes.energistics.etp.v12.protocol.store_notification.object_changed import (
    ObjectChanged,
)
from etptypes.energistics.etp.v12.protocol.store_notification.object_deleted import (
    ObjectDeleted,
)
from etptypes.energistics.etp.v12.protocol.store_notification.subscribe_notifications import (
    SubscribeNotifications,
)
from etptypes.energistics.etp.v12.protocol.store_notification.subscribe_notifications_response import (
    SubscribeNotificationsResponse,
)
from etptypes.energistics.etp.v12.protocol.store_notification.subscription_ended import (
    SubscriptionEnded,
)
from etptypes.energistics.etp.v12.protocol.store_notification.unsolicited_store_notifications import (
    UnsolicitedStoreNotifications,
)
from etptypes.energistics.etp.v12.protocol.store_notification.unsubscribe_notifications import (
    UnsubscribeNotifications,
)

from etpproto.client_info import ClientInfo
from etpproto.error import InvalidMessageTypeError, NotSupportedError
from etpproto.messages import Message
from etpproto.utils import snake_case


class StoreNotificationHandler(Protocol):
    async def on_chunk(
        self,
        msg: Chunk,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_object_access_revoked(
        self,
        msg: ObjectAccessRevoked,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_object_active_status_changed(
        self,
        msg: ObjectActiveStatusChanged,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_object_changed(
        self,
        msg: ObjectChanged,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_object_deleted(
        self,
        msg: ObjectDeleted,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_subscribe_notifications(
        self,
        msg: SubscribeNotifications,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_subscribe_notifications_response(
        self,
        msg: SubscribeNotificationsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_subscription_ended(
        self,
        msg: SubscriptionEnded,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_unsolicited_store_notifications(
        self,
        msg: UnsolicitedStoreNotifications,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_unsubscribe_notifications(
        self,
        msg: UnsubscribeNotifications,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def handle_message(
        self,
        etp_object: object,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        handling_func = getattr(
            self, "on_" + snake_case(type(etp_object).__name__)
        )
        if handling_func is not None:
            async for handled in handling_func(
                msg=etp_object,
                msg_header=msg_header,
                client_info=client_info,
            ):
                yield handled

        else:
            raise InvalidMessageTypeError()



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

import sys
from dataclasses import dataclass, field
from typing import ClassVar, List, Optional, Union


@dataclass
class EndpointCapabilityKind:

    _default: ClassVar[Optional[Union[int, str, float, bool]]] = None
    _min: ClassVar = None
    _max: ClassVar = None
    _unit: ClassVar[Optional[str]] = None


@dataclass
class ActiveTimeoutPeriod(EndpointCapabilityKind):
    """
    The minimum time period in seconds that a store keeps the active status (activeStatus field in ETP) for a data object as “active”, after the most recent update causing the data object’s active status to be set to true. For growing data objects, this is any change to its parts. For channels, this is any change to its data points.
    This capability can be set for an endpoint and/or for a data object. A data object-specific value overrides an endpoint-specific value.
    """

    _default = 3600
    _min = 60
    _max = None
    _unit = "seconds, <count of seconds>"
    value: Optional[int] = field(default=3600)


@dataclass
class AuthorizationDetails(EndpointCapabilityKind):
    """
    1. Contains an ArrayOfString with WWW-Authenticate style challenges.
    2. To support the required authorization workflow (to enable an endpoint to acquire an access token with the necessary scope from the designated authorization server), the AuthorizationDetails endpoint capability MUST include at least one challenge with the Bearer scheme which must include the ‘authz_server' and ‘scope’ parameters.
    a. The 'authz_server' parameter MUST be a URI for an authorization server to enable the endpoint to acquire any other needed metadata about the authorization server using OpenID Connect Discovery.
    3. An ETP server MUST have the AuthorizationDetails endpoint capability, which must meet the requirements of Point 2 above.
    4. If an ETP client does NOT need to authorize ETP servers, it MAY omit the AuthorizationDetails.
    """

    _default = None
    _min = None
    _max = None
    _unit = None
    value: Optional[List[str]] = field(default=None)


@dataclass
class ChangePropagationPeriod(EndpointCapabilityKind):
    """
    The maximum time period in seconds--under normal operation on an uncongested session--for these conditions:
        > after a change in an endpoint before that endpoint sends a change notification covering the change to any subscribed endpoint in any session.
        > if the change was the result of a message WITHOUT a positive response, it is the maximum time until the change is reflected in read operations in any session.
        > If the change was the result of a message WITH a positive response, it is the maximum time until the change is reflected in sessions other than the session where the change was made. RECOMMENDATION: Set as short as possible (i.e. a few seconds).
    """

    _default = 5
    _min = 1
    _max = 600
    _unit = "seconds, <count of seconds>"
    value: Optional[int] = field(default=5)


@dataclass
class ChangeRetentionPeriod(EndpointCapabilityKind):
    """
    The minimum time period in seconds that a store retains the Canonical URI of a deleted data object and any change annotations for channels and growing objects. RECOMMENDATION: This period should be as long as is feasible in an implementation. When the period is shorter, the risk is that additional data will need to be transmitted to recover from outages, leading to higher initial load on sessions.
    """

    _default = 86400
    _min = 86400
    _max = None
    _unit = "seconds, <count of seconds>"
    value: Optional[int] = field(default=86400)


@dataclass
class MaxConcurrentMultipart(EndpointCapabilityKind):
    """
    The maximum count of multipart messages allowed in parallel, in a single protocol, from one endpoint to another. The limit applies separately to each protocol, and separately from client to server and from server to client. The limit for an endpoint applies to the multipart messages that the endpoint can receive.
    EXAMPLE: If an endpoint's MaxConcurrentMultipart is 5, then it can receive 5 messages--each with any number of parts--at one time, in Store (Protocol 4) and another 5 messages in process in Discovery (Protocol 3). In Discovery (Protocol 3), this could be the multipart responses to 5 distinct GetResources request messages.
    """

    _default = None
    _min = 1
    _max = None
    _unit = "count, <count of messages>"
    value: Optional[int] = field(default=None)


@dataclass
class MaxDataObjectSize(EndpointCapabilityKind):
    """
    The maximum size in bytes of a data object allowed in a complete multipart message. Size in bytes is the size in bytes of the uncompressed string representation of the data object in the format in which it is sent or received.
    This capability can be set for an endpoint, a protocol, and/or a data object. If set for all three, here is how they generally work:
        > An object-specific value overrides an endpoint-specific value.
        > A protocol-specific value can further lower (but NOT raise) the limit for the protocol.
    EXAMPLE: A store may wish to generally support sending and receiving any data object that is one megabyte or less with the exceptions of Wells that are 100 kilobytes or less and attachments that are 5 megabytes or less. A store may further wish to limit the size of any data object sent as part of a notification in StoreNotification (Protocol 5) to 256 kilobytes.
    """

    _default = None
    _min = 100000
    _max = None
    _unit = "bytes"
    value: Optional[int] = field(default=None)


@dataclass
class MaxSessionClientCount(EndpointCapabilityKind):
    """
    The maximum count of concurrent ETP sessions that may be established for a given endpoint, by a specific client. If possible, the determination of whether this limit is exceeded should be made at the time of receiving the HTTP WebSocket upgrade or connect request based on the authorization details provided with the request. At the latest, it should be based on an authorized RequestSession message.
    """

    _default = None
    _min = 2
    _max = None
    _unit = "count, <count of sessions>"
    value: Optional[int] = field(default=None)


@dataclass
class MaxPartSize(EndpointCapabilityKind):
    """
    The The maximum size in bytes of each data object part allowed in a standalone message or a complete multipart message. Size in bytes is the total size in bytes of the uncompressed string representation of the data object part in the format in which it is sent or received.
    """

    _default = None
    _min = 100000
    _max = None
    _unit = "bytes, <number of bytes>"
    value: Optional[int] = field(default=None)


@dataclass
class MaxSessionGlobalCount(EndpointCapabilityKind):
    """
    The maximum count of concurrent ETP sessions that may be established for a given endpoint across all clients. The determination of whether this limit is exceeded should be made at the time of receiving the HTTP WebSocket upgrade or connect request. NOTE: Exposing this information may have security implications, so it should only be exposed if an implementation is comfortable with any potential associated risks.
    """

    _default = None
    _min = 2
    _max = None
    _unit = "count, <count of sessions>"
    value: Optional[int] = field(default=None)


@dataclass
class MaxWebSocketFramePayloadSize(EndpointCapabilityKind):
    """
    The maximum size in bytes allowed for a single WebSocket frame payload. The limit to use during a session is the smaller of the client's and the server's value for MaxWebSocketFramePayloadSize, which should be determined by the limits imposed by the WebSocket library used by each endpoint.
    """

    _default = None
    _min = None
    _max = None
    _unit = "bytes, <number of bytes>"
    value: Optional[int] = field(default=None)


@dataclass
class MaxWebSocketMessagePayloadSize(EndpointCapabilityKind):
    """
    The maximum size in bytes allowed for a complete WebSocket message payload, which is composed of one or more WebSocket frames. The limit to use during a session is the smaller of the client's and the server's value for MaxWebSocketMessagePayloadSize, which should be determined by the limits imposed by the WebSocket library used by each endpoint.
    """

    _default = None
    _min = None
    _max = None
    _unit = "bytes, <number of bytes>"
    value: Optional[int] = field(default=None)


@dataclass
class MultipartMessageTimeoutPeriod(EndpointCapabilityKind):
    """
    The maximum time period in seconds--under normal operation on an uncongested session--allowed between subsequent messages in the SAME multipart request or response. The period is measured as the time between when each message has been fully sent or received via the WebSocket.
    """

    _default = None
    _min = 60
    _max = None
    _unit = "seconds, <count of seconds>"
    value: Optional[int] = field(default=None)


@dataclass
class ResponseTimeoutPeriod(EndpointCapabilityKind):
    """
    The maximum time period in seconds allowed between a request and the standalone response message or the first message in the multipart response message. The period is measured as the time between when the request message has been successfully sent via the WebSocket and when the first or only response message has been fully received via the WebSocket. When calculating this period, any Acknowledge messages or empty placeholder responses are ignored EXCEPT where these are the only and final response(s) to the request.
    """

    _default = 300
    _min = 60
    _max = None
    _unit = "seconds, <count of seconds>"
    value: Optional[int] = field(default=300)


@dataclass
class RequestSessionTimeoutPeriod(EndpointCapabilityKind):
    """
    The maximum time period in seconds a server will wait to receive a RequestSession message from a client after the WebSocket connection has been established.
    """

    _default = 45
    _min = 5
    _max = None
    _unit = "seconds, <count of seconds>"
    value: Optional[int] = field(default=45)


@dataclass
class SessionEstablishmentTimeoutPeriod(EndpointCapabilityKind):
    """
    The maximum time period in seconds a client or server will wait for a valid ETP session to be established.
    For a server:
        > A valid session is established when it sends an OpenSession message to the client, which indicates a session has been successfully established.
        > The time period starts when it receives the initial RequestSession message from the client.
    For a client:
        > A valid session is established when it receives an OpenSession message from the server.
        > The time period starts when it sends the initial RequestSession message to the server.
    """

    _default = 60
    _min = 5
    _max = None
    _unit = "seconds, <count of seconds>"
    value: Optional[int] = field(default=60)


@dataclass
class SupportsAlternateRequestUris(EndpointCapabilityKind):
    """
    Indicates whether an endpoint supports alternate URI formats--beyond the canonical Energistics URIs, which MUST be supported for requests.
    """

    _default = False
    _unit = None
    value: Optional[bool] = field(default=False)


@dataclass
class SupportsMessageHeaderExtensions(EndpointCapabilityKind):
    """
    Indicates whether an endpoint supports message header extensions. For more information about message header extensions and their use, see Section 3.6.2.
    """

    _default = False
    _unit = None
    value: Optional[bool] = field(default=False)


def kind_from_name(classname):
    return getattr(sys.modules[__name__], classname)



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

from typing import AsyncGenerator, Optional, Protocol, Union

from etptypes.energistics.etp.v12.datatypes.message_header import MessageHeader
from etptypes.energistics.etp.v12.protocol.data_array.get_data_array_metadata import (
    GetDataArrayMetadata,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_array_metadata_response import (
    GetDataArrayMetadataResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_arrays import (
    GetDataArrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_arrays_response import (
    GetDataArraysResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_subarrays import (
    GetDataSubarrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_subarrays_response import (
    GetDataSubarraysResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_arrays import (
    PutDataArrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_arrays_response import (
    PutDataArraysResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_subarrays import (
    PutDataSubarrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_subarrays_response import (
    PutDataSubarraysResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_uninitialized_data_arrays import (
    PutUninitializedDataArrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_uninitialized_data_arrays_response import (
    PutUninitializedDataArraysResponse,
)

from etpproto.client_info import ClientInfo
from etpproto.error import InvalidMessageTypeError, NotSupportedError
from etpproto.messages import Message
from etpproto.utils import snake_case


class DataArrayHandler(Protocol):
    async def on_get_data_array_metadata(
        self,
        msg: GetDataArrayMetadata,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_data_array_metadata_response(
        self,
        msg: GetDataArrayMetadataResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_data_arrays(
        self,
        msg: GetDataArrays,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_data_arrays_response(
        self,
        msg: GetDataArraysResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_data_subarrays(
        self,
        msg: GetDataSubarrays,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_data_subarrays_response(
        self,
        msg: GetDataSubarraysResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_data_arrays(
        self,
        msg: PutDataArrays,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_data_arrays_response(
        self,
        msg: PutDataArraysResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_data_subarrays(
        self,
        msg: PutDataSubarrays,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_data_subarrays_response(
        self,
        msg: PutDataSubarraysResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_uninitialized_data_arrays(
        self,
        msg: PutUninitializedDataArrays,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_uninitialized_data_arrays_response(
        self,
        msg: PutUninitializedDataArraysResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def handle_message(
        self,
        etp_object: object,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        handling_func = getattr(
            self, "on_" + snake_case(type(etp_object).__name__)
        )
        if handling_func is not None:
            async for handled in handling_func(
                msg=etp_object,
                msg_header=msg_header,
                client_info=client_info,
            ):
                yield handled

        else:
            raise InvalidMessageTypeError()



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging

from dataclasses import dataclass, field
from enum import Enum
from typing import (
    AsyncGenerator,
    Callable,
    ClassVar,
    Final,
    List,
    Dict,
    Optional,
    Tuple,
    Type,
    Union,
)

from etptypes.energistics.etp.v12.datatypes.server_capabilities import (
    ServerCapabilities,
)
from etptypes.energistics.etp.v12.datatypes.supported_protocol import (
    SupportedProtocol,
)
from etptypes.energistics.etp.v12.protocol.core.acknowledge import Acknowledge
from etptypes.energistics.etp.v12.protocol.core.close_session import (
    CloseSession,
)
from etptypes.energistics.etp.v12.protocol.core.authorize_response import (
    AuthorizeResponse,
)
from etptypes.energistics.etp.v12.protocol.core.authorize import Authorize
from etptypes.energistics.etp.v12.protocol.core.open_session import OpenSession
from etptypes.energistics.etp.v12.protocol.core.request_session import (
    RequestSession,
)
from etptypes.energistics.etp.v12.datatypes.message_header import MessageHeader

from etpproto.client_info import ClientInfo
from etpproto.error import (
    ETPError,
    InvalidMessageError,
    UnsupportedProtocolError,
    NotSupportedError,
    InvalidStateError,
    AuthorizationRequired,
)
from etpproto.messages import Message
from etpproto.utils import ProtocolDict, get_all_etp_protocol_classes

# import time


class Protocol:
    async def handle_message(
        self,
        etp_object: object,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )


class ConnectionType(Enum):
    """An enumeration of connection types."""

    #: This connection will act as client and talk to a remote server
    CLIENT = 1

    #: This connection will act as as server and waits for client connections
    SERVER = 2


class CommunicationProtocol(Enum):
    """
    ETP Specification, Section 2 - ETP Published Protocols
    """

    #: Creates and manages ETP sessions.
    CORE = 0
    #: Provides "simple streamer" functionality, for example, a sensor streaming data.
    CHANNEL_STREAMING = 1
    #: Gets channel data from a store in "rows".
    CHANNEL_DATA_FRAME = 2
    #: Enables store customers to enumerate and understand the contents of a store of data objects.
    DISCOVERY = 3
    #: Performs CRUD operations (create, retrieve, update and delete) on data objects in a store.
    STORE = 4
    #: Allows store customers to receive notification of changes to data objects in the store in an event-driven manner, resulting from operations in Protocol 4.
    STORE_NOTIFICATION = 5
    #: Manages the growing parts of data objects that are index-based (i.e., time and depth) other than channels.
    GROWING_OBJECT = 6
    #: Allows a store customer to receive notifications of changes to the growing parts of growing data objects in a store, in an event-driven manner, resulting from operations in Protocol 6.
    GROWING_OBJECT_NOTIFICATION = 7
    #: Transfers large, binary arrays of heterogeneous data values, which Energistics domain standards typically store using HDF5.
    DATA_ARRAY = 9
    #: Query behavior appended to discovery functionality (which is defined in Protocol 3).
    DISCOVERY_QUERY = 13
    #: Query behavior appended to store/CRUD functionality (which is defined in Protocol 4).
    STORE_QUERY = 14
    #: Query behavior appended to growing object behavior (which is defined in Protocol 6).
    GROWING_OBJECT_QUERY = 16
    #: Handles messages associate with software application transactions, for example, end messages for applications that may have long, complex transactions (typically associated with earth modeling/RESQML).
    TRANSACTION = 18
    #: Provides standard publish/subscribe behavior for customers to connect to a store (server) and receive new channel data as available (streaming).
    CHANNEL_SUBSCRIBE = 21
    #: Enables one server (with the ETP customer role) to push (stream) data to another server (with the ETP store role).
    CHANNEL_DATALOAD = 22
    #: Used to discover dataspaces in a store. After discovering dataspaces, use Discovery (Protocol 3) to discover objects in the store.
    DATASPACE = 24
    #: Enables store customers to discover a store's data model, to dynamically understand what object types are possible in the store at a given location in the data model (though the store may have no data for these object types), without prior knowledge of the overall data model and graph connectivity.
    SUPPORTED_TYPES = 25
    #: In ETP v1.1, this protocol was published as Protocol 8. It is now a custom protocol published by an Energistics member company. .
    WITSML_SOAP = 2000


@dataclass
class ETPConnection:
    """
    Main class to start an ETP Connection.
    :cvar generic_transition_table:
    :ivar server_capabilities: The capabilities of the current ETP connection (client or server)
    :ivar transition_table: The table that maps a communication protocol to an actual protocol implementation
    :ivar is_connected:
    :ivar chunk_msg_cache: Mapping a msgId to list of partial msg (). Is is ONLY used for RECIEVED messages
    """

    SUB_PROTOCOL: Final[str] = "etp12.energistics.org"

    generic_transition_table: ClassVar[
        ProtocolDict
    ] = get_all_etp_protocol_classes()

    transition_table: ClassVar[Dict[CommunicationProtocol, Protocol]] = {}

    server_capabilities: Optional[ServerCapabilities] = field(default=None)

    is_connected: bool = field(default=False)

    auth_required: bool = field(default=False)

    message_id: int = field(default=1)

    #    ______           __                                      __             ________                __
    #   / ____/___ ______/ /_  ___     ____  ____  __  _______   / /__  _____   / ____/ /_  __  ______  / /_______   ____ ___  ___  ______________ _____ ____
    #  / /   / __ `/ ___/ __ \/ _ \   / __ \/ __ \/ / / / ___/  / / _ \/ ___/  / /   / __ \/ / / / __ \/ //_/ ___/  / __ `__ \/ _ \/ ___/ ___/ __ `/ __ `/ _ \
    # / /___/ /_/ / /__/ / / /  __/  / /_/ / /_/ / /_/ / /     / /  __(__  )  / /___/ / / / /_/ / / / / ,< (__  )  / / / / / /  __(__  |__  ) /_/ / /_/ /  __/
    # \____/\__,_/\___/_/ /_/\___/  / .___/\____/\__,_/_/     /_/\___/____/   \____/_/ /_/\__,_/_/ /_/_/|_/____/  /_/ /_/ /_/\___/____/____/\__,_/\__, /\___/
    #                              /_/                                                                                                           /____/

    chunk_msg_cache: Dict[int, List[Message]] = field(
        default_factory=lambda: {}
    )

    error_msg_cache: Dict[int, List[Optional[Message]]] = field(
        default_factory=lambda: {}
    )

    client_info: ClientInfo = field(default=ClientInfo())

    connection_type: ConnectionType = field(default=ConnectionType.SERVER)
    # TODO : last msg recieve date
    # TODO : max timeout before disconnecting

    def __post_init__(self):
        if self.connection_type == ConnectionType.SERVER:
            self.message_id = 1
        elif self.connection_type == ConnectionType.CLIENT:
            self.message_id = 2
            self.auth_required = False  # auth is only required on server side

    def _handle_answer_and_error(
        self,
        msg: Optional[Union[Message, ETPError]],
        req_msg: Message,
        request_msg_id: int,
    ) -> Optional[Message]:
        # Si on a repondu un message c'est que tout s'est bien passé
        if msg:
            if isinstance(msg, Message):
                msg.header.message_id = self.consume_msg_id()
                msg.set_final_msg(True)
                return msg

            # conversion de l'erreur en msg etp si besoin
            elif isinstance(msg, ETPError):
                return msg.to_etp_message(request_msg_id)
        return None

    async def _handle_message_generator(
        self, etp_input_msg: Optional[Message]
    ) -> AsyncGenerator[Optional[Message], None]:
        if (
            etp_input_msg is not None and etp_input_msg.header is not None
        ):  # si pas un message none
            if not self.auth_required or (
                self.client_info is not None
                and (
                    self.client_info.authenticated
                    or (
                        isinstance(etp_input_msg.body, Authorize)
                        or isinstance(etp_input_msg.body, AuthorizeResponse)
                    )
                )
            ):
                if (
                    # isinstance(etp_input_msg.body, RequestSession)
                    # or isinstance(etp_input_msg.body, OpenSession)
                    etp_input_msg.header.protocol
                    == CommunicationProtocol.CORE.value
                    or self.is_connected
                ):
                    current_msg_id = etp_input_msg.header.message_id

                    # if requires acknowledge :
                    if (
                        etp_input_msg.is_asking_acknowledge()
                        and not isinstance(etp_input_msg.body, Acknowledge)
                    ):
                        yield Message.get_object_message(
                            Acknowledge(),
                            correlation_id=current_msg_id,
                            msg_id=self.consume_msg_id(),
                        )
                        # time.sleep(3)

                    # only if the user is connected or request for an OpenSession or if the message is not the full message

                    if self.is_connected and isinstance(
                        etp_input_msg.body, CloseSession
                    ):
                        logging.debug(
                            f"{self.client_info.ip} : CloseSession recieved"
                        )
                        self.is_connected = False
                    else:
                        # Test if it is an Open/Request session
                        if (
                            isinstance(etp_input_msg.body, RequestSession)
                            and self.connection_type == ConnectionType.SERVER
                        ) or (
                            isinstance(etp_input_msg.body, OpenSession)
                            and self.connection_type == ConnectionType.CLIENT
                        ):
                            self.is_connected = True
                            self.client_info.negotiate(etp_input_msg.body)

                        # logging.debug(etp_input_msg, etp_input_msg.is_chunk_msg(), etp_input_msg.is_chunk_msg_referencer())
                        # On test si c'est un message de BLOB qu'il faut mettre en cache :
                        if etp_input_msg.is_multipart_msg() and (
                            etp_input_msg.is_chunk_msg()
                            or etp_input_msg.is_chunk_msg_referencer()
                        ):
                            cache_id = (
                                etp_input_msg.header.correlation_id
                                if etp_input_msg.header.correlation_id != 0
                                else etp_input_msg.header.message_id
                            )
                            if cache_id not in self.chunk_msg_cache:
                                self.chunk_msg_cache[cache_id] = []
                            self.chunk_msg_cache[cache_id].append(
                                etp_input_msg
                            )

                            # si final on rassemble et on handle.
                            if etp_input_msg.is_final_msg():
                                logging.debug(
                                    f"Reassemble chunks :{self.chunk_msg_cache[cache_id]}",
                                )
                                try:
                                    async for msg in self._handle_message_generator(
                                        Message.reassemble_chunk(
                                            self.chunk_msg_cache[cache_id]
                                        )
                                    ):
                                        if msg is not None:
                                            yield msg
                                        else:
                                            if (
                                                cache_id
                                                not in self.error_msg_cache
                                            ):
                                                self.error_msg_cache[
                                                    cache_id
                                                ] = []
                                            self.error_msg_cache[
                                                cache_id
                                            ].append(
                                                InvalidMessageError().to_etp_message(
                                                    msg_id=self.consume_msg_id()
                                                )
                                            )

                                except Exception as e:
                                    logging.error(
                                        f"{self.client_info.ip}: _SERVER_ not handled exception",
                                    )
                                    raise e

                                if cache_id in self.error_msg_cache:
                                    for err_msg in self.error_msg_cache[
                                        cache_id
                                    ]:
                                        if err_msg is not None:
                                            yield err_msg
                                    self.error_msg_cache.pop(cache_id)

                                if cache_id in self.chunk_msg_cache:
                                    self.chunk_msg_cache.pop(cache_id)

                        else:  # ce n'est pas un message envoye en chunks
                            # now try to have an answer
                            try:
                                # Test si le protocol est supporte par le serveur
                                if (
                                    CommunicationProtocol(
                                        etp_input_msg.header.protocol
                                    )
                                    in self.transition_table
                                ):
                                    # demande la reponse au protocols du serveur
                                    try:
                                        async for handled in self.transition_table[
                                            CommunicationProtocol(
                                                etp_input_msg.header.protocol
                                            )
                                        ].handle_message(
                                            etp_object=etp_input_msg.body,
                                            msg_header=etp_input_msg.header,
                                            client_info=self.client_info,
                                        ):
                                            yield self._handle_answer_and_error(
                                                msg=handled,
                                                req_msg=etp_input_msg,
                                                request_msg_id=current_msg_id,
                                            )
                                    except ETPError as exp_invalid_msg_type:
                                        yield exp_invalid_msg_type.to_etp_message(
                                            msg_id=self.consume_msg_id(),
                                            correlation_id=current_msg_id,
                                        )
                                else:
                                    logging.debug(
                                        f"{self.client_info.ip} : #handle_msg : unkown protocol id : {str(etp_input_msg.header.protocol)}"
                                    )
                                    raise UnsupportedProtocolError(
                                        etp_input_msg.header.protocol
                                    )
                            except ETPError as etp_err:
                                logging.error(
                                    f"{self.client_info.ip}: _SERVER_ internal error : {etp_err}"
                                )
                                yield self._handle_answer_and_error(
                                    msg=etp_err.to_etp_message(
                                        msg_id=self.consume_msg_id(),
                                        correlation_id=current_msg_id,
                                    ),
                                    req_msg=etp_input_msg,
                                    request_msg_id=current_msg_id,
                                )
                            except Exception as e:
                                logging.error(
                                    f"{self.client_info.ip}: _SERVER_ not handled exception",
                                )
                                raise e
                else:  # not connected
                    yield InvalidStateError().to_etp_message(
                        msg_id=self.consume_msg_id()
                    )
            else:  # not authenticated
                yield AuthorizationRequired().to_etp_message(
                    msg_id=self.consume_msg_id()
                )
        else:  # null message
            yield InvalidMessageError().to_etp_message(
                msg_id=self.consume_msg_id()
            )

    async def send_msg_and_error_generator(
        self, msg: Message, err_msg: Message
    ) -> AsyncGenerator[Tuple[int, bytes], None]:
        current_msg_id = self.consume_msg_id()

        if msg:
            # Attention, le flag (FIN bit) ne doit etre mis que sur le dernier msg
            msg.header.message_flags = msg.header.message_flags | 0x02
            msg.header.message_id = current_msg_id

            # err_msg.header.message_id = self.message_id
            # self.message_id += 2

            async for msg_part in msg.encode_message_generator(
                self.client_info.getCapability(
                    "MaxWebSocketMessagePayloadSize"
                ),
                self,
            ):
                yield (current_msg_id, msg_part)

            if err_msg:
                async for msg_part in err_msg.encode_message_generator(
                    self.client_info.getCapability(
                        "MaxWebSocketMessagePayloadSize"
                    ),
                    self,
                ):
                    yield (current_msg_id, msg_part)

        elif err_msg:
            # logging.debug(f"{self.client_info.ip} : Encoding error")
            async for msg_part in err_msg.encode_message_generator(
                self.client_info.getCapability(
                    "MaxWebSocketMessagePayloadSize"
                ),
                self,
            ):
                yield (current_msg_id, msg_part)

    async def handle_bytes_generator(
        self, msg_data: bytes
    ) -> AsyncGenerator[bytes, None]:
        """
        Returns a generator of binary messages to send
        """

        etp_input_msg = Message.decode_binary_message(
            msg_data, ETPConnection.generic_transition_table
        )
        logging.debug(f"### MSG {etp_input_msg}")

        async for msg in self._handle_message_generator(etp_input_msg):
            if msg is not None:
                async for msg_part in msg.encode_message_generator(
                    self.client_info.getCapability(
                        "MaxWebSocketMessagePayloadSize"
                    ),
                    self,
                ):
                    yield msg_part

    def consume_msg_id(self):
        tmp_msg_id = self.message_id
        self.message_id += 2
        return tmp_msg_id

    @classmethod
    def on(
        cls: Type[ETPConnection], protocol: CommunicationProtocol
    ) -> Callable[[Type[Protocol]], Type[Protocol]]:
        """Should only be used to decorate classes."""

        def decorate(cls_protocol: Type[Protocol]) -> Type[Protocol]:
            cls.transition_table[protocol] = cls_protocol()
            return cls_protocol

        return decorate

    @classmethod
    def dec_server_capabilities(cls: Type[ETPConnection]):
        """Should only be used to decorate classes."""

        def decorate(compute_capability):
            cls.server_capabilities = compute_capability(
                cls.get_supported_protocol_list
            )
            return compute_capability

        return decorate

    @classmethod
    def get_supported_protocol_list(
        cls: Type[ETPConnection],
    ) -> List[SupportedProtocol]:
        supported_protocols: List[SupportedProtocol] = []
        for protocol in cls.transition_table:
            if protocol.value != CommunicationProtocol.CORE:
                supported_protocols.append(
                    SupportedProtocol(
                        protocol=protocol.value,
                        protocol_version={
                            "major": 1,
                            "minor": 2,
                            "patch": 0,
                            "revision": 0,
                        },
                        role="server",
                        protocol_capabilities={},
                    )
                )
        return supported_protocols



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

from typing import AsyncGenerator, Optional, Protocol, Union

from etptypes.energistics.etp.v12.datatypes.message_header import MessageHeader
from etptypes.energistics.etp.v12.protocol.growing_object_notification.part_subscription_ended import (
    PartSubscriptionEnded,
)
from etptypes.energistics.etp.v12.protocol.growing_object_notification.parts_changed import (
    PartsChanged,
)
from etptypes.energistics.etp.v12.protocol.growing_object_notification.parts_deleted import (
    PartsDeleted,
)
from etptypes.energistics.etp.v12.protocol.growing_object_notification.parts_replaced_by_range import (
    PartsReplacedByRange,
)
from etptypes.energistics.etp.v12.protocol.growing_object_notification.subscribe_part_notifications import (
    SubscribePartNotifications,
)
from etptypes.energistics.etp.v12.protocol.growing_object_notification.subscribe_part_notifications_response import (
    SubscribePartNotificationsResponse,
)
from etptypes.energistics.etp.v12.protocol.growing_object_notification.unsolicited_part_notifications import (
    UnsolicitedPartNotifications,
)
from etptypes.energistics.etp.v12.protocol.growing_object_notification.unsubscribe_part_notification import (
    UnsubscribePartNotification,
)

from etpproto.client_info import ClientInfo
from etpproto.error import InvalidMessageTypeError, NotSupportedError
from etpproto.messages import Message
from etpproto.utils import snake_case


class GrowingObjectNotificationHandler(Protocol):
    async def on_part_subscription_ended(
        self,
        msg: PartSubscriptionEnded,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_parts_changed(
        self,
        msg: PartsChanged,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_parts_deleted(
        self,
        msg: PartsDeleted,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_parts_replaced_by_range(
        self,
        msg: PartsReplacedByRange,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_subscribe_part_notifications(
        self,
        msg: SubscribePartNotifications,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_subscribe_part_notifications_response(
        self,
        msg: SubscribePartNotificationsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_unsolicited_part_notifications(
        self,
        msg: UnsolicitedPartNotifications,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_unsubscribe_part_notification(
        self,
        msg: UnsubscribePartNotification,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def handle_message(
        self,
        etp_object: object,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        handling_func = getattr(
            self, "on_" + snake_case(type(etp_object).__name__)
        )
        if handling_func is not None:
            async for handled in handling_func(
                msg=etp_object,
                msg_header=msg_header,
                client_info=client_info,
            ):
                yield handled

        else:
            raise InvalidMessageTypeError()



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import logging

import json
import re
import traceback
import uuid as pyUUID
from abc import ABC
from copy import deepcopy
from dataclasses import dataclass
from enum import IntFlag
from io import BytesIO
from math import ceil
from typing import AsyncGenerator, Dict, List, Optional, Tuple, Any, Union

import etptypes.energistics.etp.v12.datatypes.message_header as mh
from etptypes import ETPModel, avro_schema
from etptypes.energistics.etp.v12.datatypes.object.data_object import (
    DataObject,
)
from etptypes.energistics.etp.v12.datatypes.uuid import Uuid
from fastavro import schemaless_reader, schemaless_writer

from etpproto.utils import (
    ProtocolDict,
    concat,
    get_class_from_protocol_and_name,
    get_first_dict_attribute_name,
    get_first_list_attribute_name,
)


class MessageFlags(
    IntFlag
):  # enum.Flag class would be a better choice but doesn't work with our operations
    # None
    NONE = 0x0

    # A part of a multi-part message.
    MULTIPART = 0x1

    # The final part of a multi-part message.
    FINALPART = 0x2

    # Short-hand for both mutli-part and final part: 0x1 | 0x2
    MULTIPART_AND_FINALPART = 0x3

    # No data is available.
    NO_DATA = 0x4

    # The message body is compressed.
    COMPRESSED = 0x8

    # An Acknowledge message is requested by the sender.
    ACKNOWLEDGE = 0x10

    # The message has a header extension.
    HAS_HEADER_EXTENSION = 0x20


@dataclass
class Message(ABC):
    header: mh.MessageHeader
    body: ETPModel

    def encode_message(self) -> bytes:
        bio = BytesIO()
        if self.header:
            header_schema = json.loads(mh.avro_schema)
            schemaless_writer(
                bio, header_schema, self.header.dict(by_alias=True)
            )
        objSchema = json.loads(avro_schema(type(self.body)))
        schemaless_writer(bio, objSchema, self.body.dict(by_alias=True))

        value = bio.getvalue()

        return value

    # Encodage d'un message :
    #   Si le message depasse la taille max :
    #       - Si c'est un "Plural Message" (see. 3.7.3 de la documention etp v1.1)
    #           > Si oui on essaie de decouper la map en plusieurs messages
    #               +++> Verifier si le nb de clefs ne depassent pas le MaxResponseCount, sinon on envoie un ERESPONSECOUNT_EXCEEDED (see 3.7.3.1)
    #           > Sinon si contient un BLOB, on essaie de le couper en plusieurs messages
    #           > Sinon on envoie un message d'erreur
    #       - Sinon, on si contient un BLOB, on essaie de le couper en plusieurs messages
    #       - Sinon on envoie un message d'erreur
    # max_bytes_per_msg : negative number for infinite size
    async def encode_message_generator(  # noqa
        self, max_bytes_per_msg, connection
    ) -> AsyncGenerator[bytes, None]:
        # si requete on met le correlation_id sur le l'id du premier message
        # si self a un correlation_id alors c'est que ce n'était pas le premier
        correlation_id = (
            self.header.correlation_id
            if self.header.correlation_id != 0
            else self.header.message_id
        )
        is_a_request = (
            not type(self.body).__name__.lower().endswith("response")
        )

        from etpproto.error import ETPError, MaxSizeExceededError

        # Header encoding
        header_schema = json.loads(mh.avro_schema)
        out_h0 = BytesIO()
        if self.header:
            schemaless_writer(
                out_h0, header_schema, self.header.dict(by_alias=True)
            )

        # Body encoding
        out_body = BytesIO()
        objSchema = json.loads(avro_schema(type(self.body)))
        schemaless_writer(out_body, objSchema, self.body.dict(by_alias=True))

        # Size computation
        headerSize = int(out_h0.getbuffer().nbytes)
        bodySize = int(out_body.getbuffer().nbytes)
        try:
            if (
                max_bytes_per_msg > 0
                and headerSize + bodySize > max_bytes_per_msg
            ):
                # Message exceed the max_bytes_per_msg
                if self.is_plural_msg():
                    msg1 = deepcopy(self)
                    # get the first dict attribute
                    # split it in x factor (as the message is x times too large)
                    cuttable_attrib_name = get_first_dict_attribute_name(
                        msg1.body
                    )

                    values = []
                    if cuttable_attrib_name is not None:  # on a bien un dict
                        values = getattr(msg1.body, cuttable_attrib_name)
                    else:
                        cuttable_attrib_name = get_first_list_attribute_name(
                            msg1.body
                        )
                        values = getattr(msg1.body, cuttable_attrib_name)

                    if len(values) > 1:
                        msg2 = deepcopy(self)
                        msg2.header.message_id = connection.consume_msg_id()

                        # si msg1 etait un vrai premier message on met le suivant au bon correlation id
                        if msg1.header.correlation_id == 0:
                            msg2.header.correlation_id = msg1.header.message_id

                        if is_a_request:
                            msg2.header.correlation_id = correlation_id
                        # else : # rien a changer le correlation_id est deja sur le meme que self car c'est celui de la requete recu

                        d0: Union[list, Dict[Any, Any]] = {}
                        d1: Union[list, Dict[Any, Any]] = {}

                        if isinstance(values, dict):  # on a bien un dict
                            # splitting dict in 2 parts
                            for k, v in values.items():
                                if len(d0) < len(values) / 2:
                                    d0[k] = v
                                else:
                                    d1[k] = v
                        else:
                            d0 = []
                            d1 = []
                            # splitting list in 2 parts
                            for v in values:
                                if len(d0) < len(values) / 2:
                                    d0.append(v)
                                else:
                                    d1.append(v)

                        setattr(msg1.body, cuttable_attrib_name, d0)
                        setattr(msg2.body, cuttable_attrib_name, d1)

                        msg1.add_header_flag(MessageFlags.MULTIPART)
                        msg2.add_header_flag(MessageFlags.MULTIPART)

                        msg1.set_final_msg(False)
                        async for part in msg1.encode_message_generator(
                            max_bytes_per_msg, connection
                        ):
                            yield part
                        async for part in msg2.encode_message_generator(
                            max_bytes_per_msg, connection
                        ):
                            yield part
                    else:
                        if (
                            isinstance(values, dict)
                            and len(values) > 0
                            and self.is_chunkable()
                        ):
                            async for part in _encode_message_generator_chunk(
                                chunkable_msg=self,
                                encoded_msg_size=bodySize,
                                max_bytes_per_msg=max_bytes_per_msg,
                                connection=connection,
                            ):
                                yield part
                        else:
                            raise MaxSizeExceededError()
                else:
                    raise MaxSizeExceededError()
            else:  # Original Message doesn't exceed max_bytes_per_msg
                # self.set_final_msg()
                yield out_h0.getvalue() + out_body.getvalue()
        except ETPError as err:
            msg_err = err.to_etp_message(
                msg_id=connection.consume_msg_id(),
                correlation_id=self.header.correlation_id
                if self.header.correlation_id != 0
                else 0,
            )
            if msg_err is not None:
                msg_err.set_final_msg(True)
                async for part in msg_err.encode_message_generator(
                    -1, connection
                ):
                    yield part
            else:
                raise err

    def is_partial(self) -> bool:
        return isinstance(self.body, bytes)

    def is_final_msg(self) -> bool:
        return self.header.message_flags & MessageFlags.FINALPART != 0

    def is_multipart_msg(self) -> bool:
        return self.header.message_flags & MessageFlags.MULTIPART != 0

    def is_msg_body_compressed(self) -> bool:
        return self.header.message_flags & MessageFlags.COMPRESSED != 0

    def is_asking_acknowledge(self) -> bool:
        return self.header.message_flags & MessageFlags.ACKNOWLEDGE != 0

    def is_including_optional_extension(self) -> bool:
        return (
            self.header.message_flags & MessageFlags.HAS_HEADER_EXTENSION != 0
        )

    def set_final_msg(self, is_final: bool) -> None:
        if is_final:
            self.add_header_flag(MessageFlags.FINALPART)
        else:
            self.remove_header_flag(MessageFlags.FINALPART)

    def add_header_flag(self, msg_flag: MessageFlags) -> None:
        self.header.message_flags = self.header.message_flags | msg_flag

    def remove_header_flag(self, msg_flag: MessageFlags) -> None:
        self.header.message_flags = self.header.message_flags & ~(msg_flag)

    def is_plural_msg(self):
        if self.body:
            return (
                re.fullmatch(r".*s(Response)?$", type(self.body).__name__)
                is not None
            )
        return False

    def is_chunk_msg(self) -> bool:
        return type(self.body).__name__.lower() == "chunk"

    def is_chunkable(self) -> bool:
        return hasattr(self.body, "data_objects")

    def is_chunk_msg_referencer(self):
        """
        Returns true if the message is a "chunkable" one and doesn't contain any data, but a blobId.
        This means that data will be sent after in chunk messages
        """
        if not self.is_chunk_msg() and self.is_chunkable():
            if isinstance(self.body.data_objects, list):
                for do in self.body.data_objects:
                    if not (
                        do.blob_id is not None
                        and (do.data is None or do.data == "")
                    ):
                        return False
                return True
            else:
                for k, do in self.body.data_objects.items():
                    if not (
                        do.blob_id is not None
                        and (do.data is None or len(do.data) == 0)
                    ):
                        return False
                return True

    @classmethod
    def reassemble_chunk(
        cls, multipart_msg: List[Message]
    ) -> Optional[Message]:
        referencer: Optional[Message] = None

        # on rassemble tout dans un seul msg (concat de toutes les map)
        for msg in multipart_msg:
            if msg.is_chunk_msg_referencer():
                if referencer is None:
                    referencer = msg
                else:
                    # assert referencer.body is not None
                    referencer.body.data_objects = (  # type: ignore[attr-defined]
                        msg.body.data_objects  # type: ignore[attr-defined]
                        if referencer.body.data_objects is None  # type: ignore[attr-defined]
                        else concat(
                            referencer.body.data_objects,  # type: ignore[attr-defined]
                            msg.body.data_objects,  # type: ignore[attr-defined]
                        )
                    )

        result = None

        assert referencer is not None
        do_collection = referencer.body.data_objects  # type: ignore[attr-defined]

        # todo : faire un sort ?
        if referencer is not None and do_collection is not None:

            def inner_operate_data_object(data_object: DataObject):
                for msg in multipart_msg:
                    if (
                        msg.is_chunk_msg()
                        and msg.body.blob_id == data_object.blob_id  # type: ignore[attr-defined]
                    ):
                        data_object.data = (
                            msg.body.data  # type: ignore[attr-defined]
                            if data_object.data is None
                            or len(data_object.data) == 0
                            else data_object.data + msg.body.data  # type: ignore[attr-defined]
                        )
                    # TODO : on ne test pas le "is_final" du dernier chunk mais peut importe

                if data_object.data is not None and len(data_object.data) > 0:
                    data_object.blob_id = (
                        None  # pas de blob_id ET data en meme temps
                    )
                return data_object

            if isinstance(do_collection, list):
                for do in do_collection:
                    inner_operate_data_object(do)
            elif isinstance(do_collection, dict):
                for k, do in do_collection.items():
                    inner_operate_data_object(do)
            else:
                logging.error(
                    f"#etpproto.message@reassemble_chunk : not supported chunckable message data_objects type : {type(do_collection)}"
                )
            result = referencer

        return result

    @classmethod
    def decode_binary_message(
        cls, binary: bytes, dict_map_pro_to_class: ProtocolDict
    ) -> Optional[Message]:
        fo = BytesIO(binary)
        recMH = schemaless_reader(
            fo,
            json.loads(mh.avro_schema),
            return_record_name=True,
            return_record_name_override=True,
        )
        posAfterHeaderRead = fo.tell()

        if recMH["protocol"] >= 0:
            try:
                object_class = dict_map_pro_to_class[str(recMH["protocol"])][
                    str(recMH["messageType"])
                ]

                # logging.debug("##> len : {len(binary)} posAfterHeaderRead {posAfterHeaderRead} fotell {fo.tell()}")

                object_res = schemaless_reader(
                    fo,
                    json.loads(avro_schema(object_class)),
                    return_record_name=True,
                    return_record_name_override=True,
                )

                logging.debug(f"HEADER {recMH}")

                logging.debug(
                    f"classmethod decode_binary_message {object_res}"
                )
                logging.debug(f" ==> object_class {object_class}")

                return Message(
                    mh.MessageHeader.parse_obj(recMH),
                    object_class.parse_obj(object_res),
                )
            except Exception as e:
                logging.error(f"{e}")
                # error, now we try to read it as an error, because error has now the protocol of the message send by the client
                # try:
                object_class = dict_map_pro_to_class["0"][
                    str(recMH["messageType"])
                ]

                logging.debug(f" ==> object_class {object_class}")

                object_res = schemaless_reader(
                    fo,
                    json.loads(avro_schema(object_class)),
                    return_record_name=True,
                    return_record_name_override=True,
                )
                return Message(
                    mh.MessageHeader.parse_obj(recMH),
                    object_class.parse_obj(object_res),
                )
                # except Exception:
                #     traceback.print_exc()
                #     logging.error("### ERR : in decode_binary_message")
                #     logging.error(f"{e}")
                #     pass

        # If the message has not been read, it's should be a partial message
        fo.seek(posAfterHeaderRead)
        return None  # Message(mh.MessageHeader.parse_obj(recMH), fo.read())

    @classmethod
    def get_object_message(
        cls,
        etp_object: ETPModel,
        msg_id: int = -1,
        has_header: bool = True,
        correlation_id: int = 0,
        message_flags: int = 0,
    ) -> Optional[Message]:
        if etp_object:
            logging.debug(f"get_object_message {etp_object}")
            logging.debug(f"get_object_message {type(etp_object)}")

            objSchema = json.loads(avro_schema(type(etp_object)))

            header = None
            if has_header:
                header = mh.MessageHeader(
                    protocol=int(objSchema["protocol"]),
                    messageType=int(objSchema["messageType"]),
                    correlationId=correlation_id,
                    messageId=msg_id,
                    messageFlags=message_flags,
                )
                return Message(header, etp_object)
        return None


def header_has_flag(header: mh.MessageHeader, flag: MessageFlags):
    return header.message_flags & flag != 0


def decode_binary_message(
    binary: bytes, dict_map_pro_to_class: ProtocolDict
) -> Tuple[mh.MessageHeader, ETPModel]:
    fo = BytesIO(binary)
    recMH = schemaless_reader(
        fo,
        json.loads(mh.avro_schema),
        return_record_name=True,
        return_record_name_override=True,
    )
    object_class = dict_map_pro_to_class[str(recMH["protocol"])][
        str(recMH["messageType"])
    ]
    object_res = schemaless_reader(
        fo,
        json.loads(avro_schema(object_class)),
        return_record_name=True,
        return_record_name_override=True,
    )

    logging.debug(f"decode_binary_message {object_res}")

    return (
        mh.MessageHeader.parse_obj(recMH),
        object_class.parse_obj(object_res),
    )


# When calling thins function we are supposed to have only one entry in the data_objects map/list
async def _encode_message_generator_chunk(
    chunkable_msg: Message,
    encoded_msg_size: int,
    max_bytes_per_msg: int,
    connection,
) -> AsyncGenerator[bytes, None]:
    from etpproto.error import InternalError

    secure_size = 50  # TODO : ameliorer pour que le chunk fasse vraiment la taille max d'un message (il faudrait connaitre la taille de ce qui n'est pas binaire dans le chunk message)
    size_of_chunks = max_bytes_per_msg - secure_size

    data_objs = chunkable_msg.body.data_objects  # type: ignore[attr-defined]
    msg_was_final = chunkable_msg.is_final_msg()

    correlation_id = (
        chunkable_msg.header.correlation_id
        if chunkable_msg.header.correlation_id != 0
        else chunkable_msg.header.message_id
    )

    # else : # rien a changer le correlation_id est deja sur le meme que self car c'est celui de la requete recu

    if data_objs:  # si on a une list/Map de dataObjects
        nb_chunks = ceil(
            encoded_msg_size / size_of_chunks
        )  # substract 50 for header part (header takes 5?) and non binary part of the chunk message
        logging.debug(
            f"Size of chunks : {size_of_chunks} msgS {encoded_msg_size}"
        )
        # get the chunks class
        chunk_class = get_class_from_protocol_and_name(
            str(chunkable_msg.header.protocol),
            "chunk",
            connection.generic_transition_table,
        )

        if chunk_class is not None:
            lst_chunks = []  # all chunks of all dataObjects

            # for blob_id see 3.7.3.2 of the documentation :
            # blob_id is assign to one entire DataObject and is refered in all chunck of this dataObject

            if isinstance(data_objs, dict):
                for k, do in data_objs.items():
                    data = do.data
                    blob_id = Uuid(pyUUID.uuid4().bytes)

                    # nb_chunks = ceil(len(data)/size_of_chunks)
                    for c_i in range(nb_chunks):
                        # create chunk msg
                        chunk_msg = chunk_class(
                            blob_id=blob_id,
                            data=data[
                                c_i * size_of_chunks : c_i * size_of_chunks
                                + size_of_chunks
                            ],
                            final=False,
                        )
                        lst_chunks.append(chunk_msg)

                    if len(lst_chunks) > 0:
                        lst_chunks[-1].final = True
                    do.blob_id = blob_id
                    do.data = b""  # removing the data from the message when blob_id is populated
            elif isinstance(data_objs, list):
                # on parcourt la liste des DataObjects
                # on cree plusieurs chunk pour chacuns
                for do in data_objs:
                    data = do.data
                    blob_id = Uuid(pyUUID.uuid4().bytes)

                    # nb_chunks = ceil(len(data)/size_of_chunks)
                    logging.debug(f"Nb chunks : {nb_chunks}")
                    for c_i in range(nb_chunks):
                        # create chunk msg
                        chunk_msg = chunk_class(
                            blob_id=blob_id,
                            data=data[
                                c_i * size_of_chunks : c_i * size_of_chunks
                                + size_of_chunks
                            ],
                            final=False,
                        )
                        lst_chunks.append(chunk_msg)

                    if len(lst_chunks) > 0:
                        lst_chunks[-1].final = True
                    do.blob_id = blob_id
                    do.data = b""  # removing the data from the message when blob_id is populated

            # chunkable_msg.data_objects = data_objs  # useless ?
            # send the message
            chunkable_msg.set_final_msg(False)
            chunkable_msg.add_header_flag(MessageFlags.MULTIPART)
            async for part in chunkable_msg.encode_message_generator(
                max_bytes_per_msg, connection
            ):
                yield part

            # send chunks
            for chunk in lst_chunks[:-1]:
                current_chunk_msg = Message.get_object_message(
                    etp_object=chunk,
                    msg_id=connection.consume_msg_id(),
                    has_header=True,
                    correlation_id=correlation_id,
                    message_flags=MessageFlags.MULTIPART,
                )
                if current_chunk_msg is not None:
                    async for part in current_chunk_msg.encode_message_generator(
                        max_bytes_per_msg, connection
                    ):
                        yield part
                else:
                    raise InternalError(
                        "@Message : error during chunk creation "
                        + str(chunkable_msg.header.protocol)
                    )

            # denier chunk, mettre le final a True
            final_chunk_msg = Message.get_object_message(
                etp_object=lst_chunks[-1],
                msg_id=connection.consume_msg_id(),
                has_header=True,
                correlation_id=correlation_id,
                message_flags=MessageFlags.MULTIPART_AND_FINALPART
                if msg_was_final
                else MessageFlags.MULTIPART,
            )
            if final_chunk_msg is not None:
                async for part in final_chunk_msg.encode_message_generator(
                    max_bytes_per_msg, connection
                ):
                    yield part

                # TODO : potentielle erreur de MultipartCancelledError si le nb de reponse depasse le nombre max
            else:
                raise InternalError(
                    "@Message : error during chunk creation "
                    + str(chunkable_msg.header.protocol)
                )

        else:
            raise InternalError(
                "@Message : No chunck class found for protocol "
                + str(chunkable_msg.header.protocol)
            )
    else:
        raise InternalError(
            "@Message : No data_object found "
            + str(chunkable_msg.header.protocol)
        )
    # raise une erreur ?



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

import re
from typing import Match, Optional, Union


class Uri(object):
    def __init__(self, uri: str) -> None:
        self.raw_uri = uri


CANONICAL_DATASPACE_URIS = (
    r"^eml:\/\/\/(?:dataspace\('?(?P<dataspace>[^']*?(?:''[^']*?)*)'?\))?$"
)


class DataspaceUri(Uri):
    """
    The canonical form of the default dataspace URI is:
        eml:///
    The canonical form for all other dataspace URIs is:
        eml:///dataspace('{path}')

    - In addition to named dataspaces, all ETP stores and producers MUST support a default, nameless dataspace, which is identified by the empty string.
    - While the default dataspace MUST be supported, it MAY be empty; that is, it may not have any data objects in it.

    IMPORTANT: The default dataspace is NOT an alias for a named dataspace. It is a simplification for ETP stores and producers that do not need to support named datapsaces.
    """

    def __init__(self, uri: str) -> None:
        super(DataspaceUri, self).__init__(uri)
        p = re.compile(CANONICAL_DATASPACE_URIS)
        result = p.search(uri)
        if result is not None:
            self.dataspace = result.group("dataspace")
        else:
            raise AttributeError


# Val: je separe le pkgName car dans le WS on a un package experimental : proposal
PKG_NAMES_REGXP = "[a-zA-Z]+\\w+"  # "witsml|resqml|prodml|eml"

UUID_REGEX = r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}"
CANONICAL_DATA_OBJECT_URIS = (
    r"^eml:\/\/\/(?:dataspace\('(?P<dataspace>[^']*?(?:''[^']*?)*)'\)\/)?(?P<domain>"
    + PKG_NAMES_REGXP
    + r")(?P<domainVersion>[1-9]\d+)\.(?P<objectType>\w+)\((?:(?P<uuid>"
    + UUID_REGEX
    + r")|uuid=(?P<uuid2>"
    + UUID_REGEX
    + r"),version='(?P<version>[^']*?(?:''[^']*?)*)')\)$"
)


class DataObjectURI(Uri):
    def __init__(self, uri: str) -> None:
        super(DataObjectURI, self).__init__(uri)
        p = re.compile(CANONICAL_DATA_OBJECT_URIS)
        result = p.search(uri)
        if result is not None:
            self.dataspace = result.group("dataspace")
            self.domain = result.group("domain")
            self.domain_version = result.group("domainVersion")
            self.object_type = result.group("objectType")

            if result.group("uuid") is None:
                self.uuid = result.group("uuid2")
            else:
                self.uuid = result.group("uuid")

            self.version = result.group("version")
        else:
            raise AttributeError

    @staticmethod
    def validate(uri: str) -> bool:
        return re.match(CANONICAL_DATA_OBJECT_URIS, uri) is not None


def find_uuid(input: str) -> Optional[str]:
    p = re.compile(UUID_REGEX)
    result = p.search(input)
    if result is not None:
        return result.group() if result else None
    else:
        return None


def parse_uri(uri: str) -> Union[DataObjectURI, DataspaceUri, Uri]:
    try:
        return DataObjectURI(uri)
    except AttributeError:  # si un [..].group ne marche pas
        try:
            return DataspaceUri(uri)
        except AttributeError:
            # logging.error(e2)
            return Uri(uri)



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

import importlib
import inspect
import json
import pkgutil
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Type, Union

import etptypes.energistics.etp.v12.protocol
from etptypes import ETPModel, avro_schema


def list_classes(modulePath: str) -> List[Type[ETPModel]]:
    module = importlib.import_module(modulePath)
    classList = []
    for _, obj in inspect.getmembers(module):
        if inspect.isclass(obj):
            classList.append(obj)
    return classList


MessageDict = DefaultDict[str, Type[ETPModel]]
ProtocolDict = DefaultDict[str, MessageDict]


def get_all_etp_protocol_classes() -> ProtocolDict:
    protocolDict: ProtocolDict = defaultdict(
        lambda: defaultdict(type(ETPModel))
    )
    package = etptypes.energistics.etp.v12.protocol
    for _, modname, _ in pkgutil.walk_packages(
        path=getattr(package, "__path__"),
        prefix=package.__name__ + ".",
        onerror=lambda x: None,
    ):
        for classFound in list_classes(modname):
            try:
                schem = json.loads(avro_schema(classFound))
                protocolId = schem["protocol"]
                messageType = schem["messageType"]
                protocolDict[protocolId][messageType] = classFound
            except Exception:
                pass
    return protocolDict


def get_class_from_protocol_and_name(
    protocol: str, name: str, dict_map_pro_to_class: ProtocolDict
):
    protocol = str(protocol)
    if protocol in dict_map_pro_to_class:
        for msg_t in dict_map_pro_to_class[protocol]:
            if (
                dict_map_pro_to_class[protocol][msg_t].__name__.lower()
                == name.lower()
            ):
                return dict_map_pro_to_class[protocol][msg_t]
    return None


def get_first_dict_attribute_name(obj):
    for att in obj.__dict__:
        if type(getattr(obj, att)) == dict:
            return att
    return None


def get_first_list_attribute_name(obj):
    for att in obj.__dict__:
        if type(getattr(obj, att)) == list:
            return att
    return None


class StringType:
    UPPER = 1
    LOWER = 2
    NUMERIC = 3
    OTHER = 4


def classify(character: str) -> int:
    """String classifier."""
    if character.isupper():
        return StringType.UPPER
    if character.islower():
        return StringType.LOWER
    if character.isnumeric():
        return StringType.NUMERIC

    return StringType.OTHER


def split_words(value: str) -> List[str]:
    """Split a string on new capital letters and not alphanumeric
    characters."""
    words: List[str] = []
    buffer: List[str] = []
    previous = None

    def flush():
        if buffer:
            words.append("".join(buffer))
            buffer.clear()

    for char in value:
        tp = classify(char)
        if tp == StringType.OTHER:
            flush()
        elif not previous or tp == previous:
            buffer.append(char)
        elif tp == StringType.UPPER and previous != StringType.UPPER:
            flush()
            buffer.append(char)
        else:
            buffer.append(char)

        previous = tp

    flush()
    return words


def snake_case(value: str, **kwargs: Any) -> str:
    """Convert the given string to snake case."""
    return "_".join(map(str.lower, split_words(value)))


def pascal_case(value: str, **kwargs: Any) -> str:
    """Convert the given string to pascal case."""
    return "".join(map(str.title, split_words(value)))


def concat(a: Union[List, Dict], b: Union[List, Dict]):
    if isinstance(a, list) and isinstance(b, List):
        return a + b
    elif isinstance(a, dict) and isinstance(b, dict):
        x = a.copy()
        x.update(b)
        return x
    elif isinstance(a, list) and isinstance(b, dict):
        return a + [b[key] for key in b]



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

from typing import AsyncGenerator, Optional, Protocol, Union

from etptypes.energistics.etp.v12.datatypes.message_header import MessageHeader
from etptypes.energistics.etp.v12.protocol.channel_streaming.channel_data import (
    ChannelData,
)
from etptypes.energistics.etp.v12.protocol.channel_streaming.channel_metadata import (
    ChannelMetadata,
)
from etptypes.energistics.etp.v12.protocol.channel_streaming.start_streaming import (
    StartStreaming,
)
from etptypes.energistics.etp.v12.protocol.channel_streaming.stop_streaming import (
    StopStreaming,
)
from etptypes.energistics.etp.v12.protocol.channel_streaming.truncate_channels import (
    TruncateChannels,
)

from etpproto.client_info import ClientInfo
from etpproto.error import InvalidMessageTypeError, NotSupportedError
from etpproto.messages import Message
from etpproto.utils import snake_case


class ChannelStreamingHandler(Protocol):
    async def on_channel_data(
        self,
        msg: ChannelData,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_channel_metadata(
        self,
        msg: ChannelMetadata,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_start_streaming(
        self,
        msg: StartStreaming,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_stop_streaming(
        self,
        msg: StopStreaming,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_truncate_channels(
        self,
        msg: TruncateChannels,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def handle_message(
        self,
        etp_object: object,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        handling_func = getattr(
            self, "on_" + snake_case(type(etp_object).__name__)
        )
        if handling_func is not None:
            async for handled in handling_func(
                msg=etp_object,
                msg_header=msg_header,
                client_info=client_info,
            ):
                yield handled

        else:
            raise InvalidMessageTypeError()



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

from typing import AsyncGenerator, Optional, Protocol, Union

from etptypes.energistics.etp.v12.datatypes.message_header import MessageHeader
from etptypes.energistics.etp.v12.protocol.channel_data_load.channel_data import (
    ChannelData,
)
from etptypes.energistics.etp.v12.protocol.channel_data_load.channels_closed import (
    ChannelsClosed,
)
from etptypes.energistics.etp.v12.protocol.channel_data_load.close_channels import (
    CloseChannels,
)
from etptypes.energistics.etp.v12.protocol.channel_data_load.open_channels import (
    OpenChannels,
)
from etptypes.energistics.etp.v12.protocol.channel_data_load.open_channels_response import (
    OpenChannelsResponse,
)
from etptypes.energistics.etp.v12.protocol.channel_data_load.replace_range import (
    ReplaceRange,
)
from etptypes.energistics.etp.v12.protocol.channel_data_load.replace_range_response import (
    ReplaceRangeResponse,
)
from etptypes.energistics.etp.v12.protocol.channel_data_load.truncate_channels import (
    TruncateChannels,
)
from etptypes.energistics.etp.v12.protocol.channel_data_load.truncate_channels_response import (
    TruncateChannelsResponse,
)

from etpproto.client_info import ClientInfo
from etpproto.error import InvalidMessageTypeError, NotSupportedError
from etpproto.messages import Message
from etpproto.utils import snake_case


class ChannelDataLoadHandler(Protocol):
    async def on_channel_data(
        self,
        msg: ChannelData,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_channels_closed(
        self,
        msg: ChannelsClosed,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_close_channels(
        self,
        msg: CloseChannels,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_open_channels(
        self,
        msg: OpenChannels,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_open_channels_response(
        self,
        msg: OpenChannelsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_replace_range(
        self,
        msg: ReplaceRange,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_replace_range_response(
        self,
        msg: ReplaceRangeResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_truncate_channels(
        self,
        msg: TruncateChannels,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_truncate_channels_response(
        self,
        msg: TruncateChannelsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def handle_message(
        self,
        etp_object: object,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        handling_func = getattr(
            self, "on_" + snake_case(type(etp_object).__name__)
        )
        if handling_func is not None:
            async for handled in handling_func(
                msg=etp_object,
                msg_header=msg_header,
                client_info=client_info,
            ):
                yield handled

        else:
            raise InvalidMessageTypeError()



# Copyright (c) 2022-2023 Geosiris.
# SPDX-License-Identifier: Apache-2.0

from typing import AsyncGenerator, Optional, Protocol, Union

from etptypes.energistics.etp.v12.datatypes.message_header import MessageHeader
from etptypes.energistics.etp.v12.protocol.growing_object.delete_parts import (
    DeleteParts,
)
from etptypes.energistics.etp.v12.protocol.growing_object.delete_parts_response import (
    DeletePartsResponse,
)
from etptypes.energistics.etp.v12.protocol.growing_object.get_change_annotations import (
    GetChangeAnnotations,
)
from etptypes.energistics.etp.v12.protocol.growing_object.get_change_annotations_response import (
    GetChangeAnnotationsResponse,
)
from etptypes.energistics.etp.v12.protocol.growing_object.get_growing_data_objects_header import (
    GetGrowingDataObjectsHeader,
)
from etptypes.energistics.etp.v12.protocol.growing_object.get_growing_data_objects_header_response import (
    GetGrowingDataObjectsHeaderResponse,
)
from etptypes.energistics.etp.v12.protocol.growing_object.get_parts import (
    GetParts,
)
from etptypes.energistics.etp.v12.protocol.growing_object.get_parts_by_range import (
    GetPartsByRange,
)
from etptypes.energistics.etp.v12.protocol.growing_object.get_parts_by_range_response import (
    GetPartsByRangeResponse,
)
from etptypes.energistics.etp.v12.protocol.growing_object.get_parts_metadata import (
    GetPartsMetadata,
)
from etptypes.energistics.etp.v12.protocol.growing_object.get_parts_metadata_response import (
    GetPartsMetadataResponse,
)
from etptypes.energistics.etp.v12.protocol.growing_object.get_parts_response import (
    GetPartsResponse,
)
from etptypes.energistics.etp.v12.protocol.growing_object.put_growing_data_objects_header import (
    PutGrowingDataObjectsHeader,
)
from etptypes.energistics.etp.v12.protocol.growing_object.put_growing_data_objects_header_response import (
    PutGrowingDataObjectsHeaderResponse,
)
from etptypes.energistics.etp.v12.protocol.growing_object.put_parts import (
    PutParts,
)
from etptypes.energistics.etp.v12.protocol.growing_object.put_parts_response import (
    PutPartsResponse,
)
from etptypes.energistics.etp.v12.protocol.growing_object.replace_parts_by_range import (
    ReplacePartsByRange,
)
from etptypes.energistics.etp.v12.protocol.growing_object.replace_parts_by_range_response import (
    ReplacePartsByRangeResponse,
)

from etpproto.client_info import ClientInfo
from etpproto.error import InvalidMessageTypeError, NotSupportedError
from etpproto.messages import Message
from etpproto.utils import snake_case


class GrowingObjectHandler(Protocol):
    async def on_delete_parts(
        self,
        msg: DeleteParts,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_delete_parts_response(
        self,
        msg: DeletePartsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_change_annotations(
        self,
        msg: GetChangeAnnotations,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_change_annotations_response(
        self,
        msg: GetChangeAnnotationsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_growing_data_objects_header(
        self,
        msg: GetGrowingDataObjectsHeader,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_growing_data_objects_header_response(
        self,
        msg: GetGrowingDataObjectsHeaderResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_parts(
        self,
        msg: GetParts,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_parts_by_range(
        self,
        msg: GetPartsByRange,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_parts_by_range_response(
        self,
        msg: GetPartsByRangeResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_parts_metadata(
        self,
        msg: GetPartsMetadata,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_parts_metadata_response(
        self,
        msg: GetPartsMetadataResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_get_parts_response(
        self,
        msg: GetPartsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_growing_data_objects_header(
        self,
        msg: PutGrowingDataObjectsHeader,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_growing_data_objects_header_response(
        self,
        msg: PutGrowingDataObjectsHeaderResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_parts(
        self,
        msg: PutParts,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_put_parts_response(
        self,
        msg: PutPartsResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_replace_parts_by_range(
        self,
        msg: ReplacePartsByRange,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def on_replace_parts_by_range_response(
        self,
        msg: ReplacePartsByRangeResponse,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        yield NotSupportedError().to_etp_message(
            correlation_id=msg_header.message_id
        )

    async def handle_message(
        self,
        etp_object: object,
        msg_header: MessageHeader,
        client_info: Union[None, ClientInfo] = None,
    ) -> AsyncGenerator[Optional[Message], None]:
        handling_func = getattr(
            self, "on_" + snake_case(type(etp_object).__name__)
        )
        if handling_func is not None:
            async for handled in handling_func(
                msg=etp_object,
                msg_header=msg_header,
                client_info=client_info,
            ):
                yield handled

        else:
            raise InvalidMessageTypeError()
